import java.util.*;
import java.lang.*;
import java.awt.*;
import java.io.*;


class Track {



	/**
	 *  The Drivers associated with this Track
	 */
	private Vector drivers_;
	public Vector drivers() { return drivers_; }
	public void drivers( Vector aVector ) { drivers_ = aVector; }

	/**
	 *  The frames associated with this Track
	 */
	private Hashtable travelerFrames_;
	public Hashtable travelerFrames() { return travelerFrames_; }
	public void travelerFrames( Hashtable anHashtable ) { travelerFrames_ = anHashtable; }

	/**
	 *  The name of associated with this track
	 */
	private String name_;
	public String name() { return name_; }
	public void name( String aString ) { name_ = aString; }

	/**
	 *  The name of the light lane associated this track
	 */
	private String lightName_;
	public String lightName() { return lightName_; }
	public void lightName( String aString ) { lightName_ = aString; }

	/**
	 *  Set so others can know if the track has been exhausted
	 */
	private boolean isExhausted_;
	public boolean isExhausted() { return isExhausted_; }
	public void isExhausted( boolean aBoolean ) { isExhausted_ = aBoolean; }

	/**
	 *  The sensor associated with this track
	 */
	private Sensor sensor_;
	public Sensor sensor() { return sensor_; }
	public void sensor( Sensor aSensor ) { sensor_ = aSensor; }

	Track() {
	    this.initialize();
	}

	private void initialize() {

        this.drivers(new Vector());

	    Hashtable anHT = new Hashtable();
	    this.travelerFrames(anHT);
	}

	public TravelerFrame getAppropriateTravelerFrame(int position) {

        TravelerFrame appropriateFrame = null;

	    // truncate

	    int truncatedPosition = position / 100;
	    truncatedPosition = truncatedPosition * 100;

	    String theKey = String.valueOf(truncatedPosition);
	  //  if (this.travelerFrames().containsKey(theKey)) {
	    if (position <8000) {
	        this.isExhausted(false);
	        appropriateFrame = (TravelerFrame) this.travelerFrames().get(theKey);
	        this.computeVelocityVector(truncatedPosition );
	        return appropriateFrame;
	    }
	    else {
	        this.isExhausted(true);
	        // always return the last one
	        appropriateFrame = (TravelerFrame) this.travelerFrames().get("8000");
	        this.computeVelocityVector(8000);

	        return appropriateFrame;
	    }

	}

	public Point computeVelocityVector(int thePosition) {

	    // get the current frame position
	    String theKey = String.valueOf(thePosition);
	    TravelerFrame theFrame = (TravelerFrame) this.travelerFrames().get(theKey);
	    Point pos1 = theFrame.position();

	    // get the next frame position
	    theKey = String.valueOf(thePosition + 100);
	    theFrame = (TravelerFrame) this.travelerFrames().get(theKey);
	    Point pos2 = theFrame.position();

	    // compute the vector and return
	    Point thePoint = new Point(pos2.x - pos1.x, pos2.y - pos1.y);
	      //  System.out.println("velocity vector = " + thePoint);
	    return thePoint;
	}

    public void removeDeadDrivers() {

        Vector currentDrivers = new Vector();

        Enumeration theOC = this.drivers().elements();
        while (theOC.hasMoreElements()) {
            Driver aDriver = (Driver) theOC.nextElement();
            if (!this.isExhausted()) {
                currentDrivers.addElement(aDriver);
            }
        }
        this.drivers(currentDrivers);
      //  System.out.println("Track Driver Count = " + this.drivers().size());
    }

    public void spawnDriver(TLSView theView) {

        boolean spawnFlag = true;

        Enumeration theOC = this.getVisibleCars().elements();
        while (theOC.hasMoreElements()) {
            VisibleCar theVC = (VisibleCar) theOC.nextElement();
            if (theVC.position() < 800) {    // only allow spawn if spawning area is clear
                spawnFlag = false;
            }
        }
        if (spawnFlag) {
            Driver theDriver = new Driver(theView.trafficLiteSimulator(), this);
            theDriver.car().acceleration(5);
            this.drivers().addElement(theDriver);
        }
    }



    public void letDriversReact(TLSView theView) {

        Enumeration theOC = this.drivers().elements();
        while (theOC.hasMoreElements()) {
            Driver aDriver = (Driver) theOC.nextElement();
            aDriver.lookForSignalsAndReact(theView, this);
        }
    }

    public void sendSensorInputTo(TLSView theView) {

        Enumeration theOC = this.drivers().elements();
        while (theOC.hasMoreElements()) {
            Driver aDriver = (Driver) theOC.nextElement();
            int theCurrentPosition = aDriver.car().visibleCar().position();

            if (theCurrentPosition > this.sensor().lowLimit()
                                   & theCurrentPosition < this.sensor().highLimit()) {
                theView.sensorButtonPressed(this.sensor().name());
            }
        }
    }

    public void handVisibleCarsTo(TLSView theView) {

        Enumeration theOC = this.drivers().elements();
        while (theOC.hasMoreElements()) {
            Driver aDriver = (Driver) theOC.nextElement();
            VisibleCar aVC = aDriver.car().visibleCar();
            ((IntersectionCanvas) (theView.getWidget("intersection"))).addVisibleCar(aVC);
        }
         ((IntersectionCanvas) (theView.getWidget("intersection"))).repaint();
    }


    public synchronized void loadFrameData(String filename, int xOffset, int yOffset) {

        String aLine = null;

        try {
            DataInputStream theStream = new DataInputStream (
                    new FileInputStream (filename));

            while  ((aLine = theStream.readLine()) != null) {
                StringTokenizer t = new StringTokenizer(aLine,",\n\r");
                TravelerFrame aTF = new TravelerFrame();
                aTF.track(this);
                String theKey = t.nextToken();

                aTF.position().x = xOffset + (Integer.valueOf(t.nextToken())).intValue();
                aTF.position().y = yOffset + (Integer.valueOf(t.nextToken())).intValue();
                aTF.angle(t.nextToken());
                aTF.image(CarImageFinder.instance().findImage(this.angleToImageKey(aTF.angle())));

                aTF.createPolygon();
                this.travelerFrames().put(theKey,aTF);
            }
            theStream.close();

        }
        catch (IOException e) {System.out.println("Track File Err - ");
        }

    }

    public String angleToImageKey(String stringAngle) {
        if (stringAngle.length() == 1) {
            return ("car" + "00" + stringAngle);
        }
        if (stringAngle.length() == 2) {
            return ("car" + "0" + stringAngle);
        }
        return "car" + stringAngle;
    }

   public Vector getVisibleCars() {
        Vector theVisibleCars = new Vector();
        Enumeration theDC = this.drivers().elements();
        while (theDC.hasMoreElements()) {
            Driver aDriver = (Driver) theDC.nextElement();
            theVisibleCars.addElement(aDriver.car().visibleCar());
        }
        return theVisibleCars;
    }


}

