import java.util.*;

class TrafficLiteSimulator implements Timed {


	/**
	 */
	private TLSView view_;
	public TLSView view() { return view_; }
	public void view( TLSView aTLSView ) { view_ = aTLSView; }

    private static int BEATTIME = 300; // milliseconds was 200ms

    private Timer timer_;
    public Timer timer() { return timer_; }
    public void timer (Timer aTimer) { timer_ = aTimer; }

    private int heartbeatCount_;
    public int heartbeatCount() { return heartbeatCount_; }
    public void heartbeatCount (int aCount) { heartbeatCount_ = aCount; }

	/**
	 *  The tracks that cars may follow (Driver controls speed only)
	 */
	private Hashtable tracks_;
	public Hashtable tracks() { return tracks_; }
	public void tracks( Hashtable anHashtable ) { tracks_ = anHashtable; }

	/**
	 *  The TLCAdapter associated with this TrafficLiteSimulator
	 */
	private TLCAdapter tlcAdapter_;
	public TLCAdapter tlcAdapter() { return tlcAdapter_; }
	public void tlcAdapter( TLCAdapter aTLCAdapter ) { tlcAdapter_ = aTLCAdapter; }

	/**
	 *  The Drivers associated with this TrafficLiteSimulator
	 */
	private Vector drivers_;
	public Vector drivers() { return drivers_; }
	public void drivers( Vector aVector ) { drivers_ = aVector; }

	/**
	 *  Responsible for finding collisions when they occur
	 *  in the View
	 */
	private CollisionDetector collisionDetector_;
	public CollisionDetector collisionDetector() { return collisionDetector_; }
	public void collisionDetector( CollisionDetector aCollisionDetector ) { collisionDetector_ = aCollisionDetector; }


    TrafficLiteSimulator(TLSView aView){
        this();
        this.view( aView );
        this.timer (new Timer (this, TrafficLiteSimulator.BEATTIME)); // heartbeat
    //    this.timer().start();
    }

    TrafficLiteSimulator(){
        this.initialize();

    }
    private void initialize() {

        this.tracks(new Hashtable());
        this.initializeTracks();

        this.tlcAdapter(new TLCAdapter());
        this.tlcAdapter().trafficLiteSimulator(this); // hookup the reverse reference
        this.collisionDetector(new CollisionDetector());
        this.heartbeatCount(0);
    }

    public void initializeTracks() {

        Track aTrack = new TrackNThru();
        this.tracks().put(aTrack.name(), aTrack);

        aTrack = new TrackSThru();
        this.tracks().put(aTrack.name(), aTrack);

        aTrack = new TrackWThru();
        this.tracks().put(aTrack.name(), aTrack);

        aTrack = new TrackEThru();
        this.tracks().put(aTrack.name(), aTrack);

        aTrack = new TrackWLeft();
        this.tracks().put(aTrack.name(), aTrack);

        aTrack = new TrackELeft();
        this.tracks().put(aTrack.name(), aTrack);

        aTrack = new TrackSLeft();
        this.tracks().put(aTrack.name(), aTrack);

        aTrack = new TrackNLeft();
        this.tracks().put(aTrack.name(), aTrack);

        aTrack = new TrackNRight();
        this.tracks().put(aTrack.name(), aTrack);

        aTrack = new TrackSRight();
        this.tracks().put(aTrack.name(), aTrack);

        aTrack = new TrackWRight();
        this.tracks().put(aTrack.name(), aTrack);

        aTrack = new TrackERight();
        this.tracks().put(aTrack.name(), aTrack);
    }


    public void updateSimulator(String widgetName, String command){
        this.view().update(widgetName, command);

    }
    public void instructLight(String widgetName, String command){
        this.view().update(widgetName, command);

    }

    public void nPedLite(String s){
        this.updateView("NPedLight1",s);
        this.updateView("NPedLight2",s);
    }
    public void sPedLite(String s){
        this.updateView("SPedLight1",s);
        this.updateView("SPedLight2",s);
    }
    public void ePedLite(String s){
        this.updateView("EPedLight1",s);
        this.updateView("EPedLight2",s);
    }
    public void wPedLite(String s){
        this.updateView("WPedLight1",s);
        this.updateView("WPedLight2",s);
    }

    public void northThruLite(String s){
        this.updateView("NSLight",s);
    }
    public void northLeftLite(String s){
        this.updateView("NLLight",s);
    }
    public void southThruLite(String s){
        this.updateView("SSLight",s);
    }
    public void southLeftLite(String s){
        this.updateView("LSLight",s);
    }
    public void eastThruLite(String s){
        this.updateView("ESLight",s);
    }
    public void eastLeftLite(String s){
        this.updateView("ELLight",s);
    }
    public void westThruLite(String s){
        this.updateView("WSLight",s);
    }
    public void westLeftLite(String s){
        this.updateView("WLLight",s);
    }

    private void updateView(String widgetName, String command){
        this.view().update(widgetName, command);
    }

    public void sensorButtonPressed(String actionCommand){
     //   System.out.println("sensorButtonPressed in TrafficLiteSimulator " + actionCommand);
        if( actionCommand.equals( "ns" ) ){
            this.tlcAdapter().nThruTrig();
            return;
        }
        if( actionCommand.equals( "nl" ) ){
            this.tlcAdapter().nLeftTrig();
            return;
        }
        if( actionCommand.equals( "ss" ) ){
            this.tlcAdapter().sThruTrig();
            return;
        }
        if( actionCommand.equals( "sl" ) ){
            this.tlcAdapter().sLeftTrig();
            return;
        }
        if( actionCommand.equals( "es" ) ){
            this.tlcAdapter().eThruTrig();
            return;
        }
        if( actionCommand.equals( "el" ) ){
            this.tlcAdapter().eLeftTrig();
            return;
        }
        if( actionCommand.equals( "ws" ) ){
            this.tlcAdapter().wThruTrig();
            return;
        }
        if( actionCommand.equals( "wl" ) ){
            this.tlcAdapter().wLeftTrig();
            return;
        }
    }
    public void pedestrianButtonPressed(String actionCommand){
        if( actionCommand.equals( "nw" ) ){
            this.tlcAdapter().nwPedTrig();
            return;
        }
        if( actionCommand.equals( "ne" ) ){
            this.tlcAdapter().nePedTrig();
            return;
        }
        if( actionCommand.equals( "sw" ) ){
            this.tlcAdapter().swPedTrig();
            return;
        }
        if( actionCommand.equals( "se" ) ){
            this.tlcAdapter().sePedTrig();
            return;
        }
        if( actionCommand.equals( "wn" ) ){
            this.tlcAdapter().wnPedTrig();
            return;
        }
        if( actionCommand.equals( "ws" ) ){
            this.tlcAdapter().wsPedTrig();
            return;
        }
        if( actionCommand.equals( "en" ) ){
            this.tlcAdapter().enPedTrig();
            return;
        }
        if( actionCommand.equals( "es" ) ){
            this.tlcAdapter().esPedTrig();
            return;
        }
    }
     ////////////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////
   ////////////////////////////////////////////////////////////
  //////  this is where it all happens  //////////////////////
 ////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////
    public void tick(Timer aTimer) {
        this.detectCollisions();
        this.removeDeadDrivers();
        this.spawnNewDrivers();
        this.instructDrivers();
        this.sendSensorInputToTLC();
        this.giveCarsToIntersection();

    }
    private void detectCollisions(){
        // collect the visible cars for the collision detector

        Vector theVisibleCars = new Vector();

        Enumeration theOC = this.tracks().elements();
        while (theOC.hasMoreElements()) {
            Track aTrack = (Track) theOC.nextElement();
            Enumeration theDC = aTrack.drivers().elements();
            while (theDC.hasMoreElements()) {
                Driver aDriver = (Driver) theDC.nextElement();
                theVisibleCars.addElement(aDriver.car().visibleCar());
            }
        }

        this.collisionDetector().detectCollisionsFor(theVisibleCars);

    }
    private void removeDeadDrivers(){
        // remove dead drivers

         Enumeration theOC = this.tracks().elements();
        while (theOC.hasMoreElements()) {
            Track aTrack = (Track) theOC.nextElement();
            aTrack.removeDeadDrivers();
        }
    }
    private void spawnNewDrivers(){
        // spawn new drivers

        Enumeration theOC = this.tracks().elements();
        while (theOC.hasMoreElements()) {
            Track aTrack = (Track) theOC.nextElement();
            if (Math.random() > 0.996) {  //0.996
                aTrack.spawnDriver(this.view());
            }
        }
    }
    private void instructDrivers(){

        // let drivers respond to themselves

        // north thru and north right
        Vector group1 = new Vector();
        Enumeration enum1 = ((Track) this.tracks().get("nThru")).drivers().elements();
        while(enum1.hasMoreElements()){
            Driver driver1 = (Driver) enum1.nextElement();
            group1.addElement(driver1);
        }
        Enumeration enum2 = ((Track) this.tracks().get("nRight")).drivers().elements();
        while(enum2.hasMoreElements()){
            Driver driver1 = (Driver) enum2.nextElement();

            group1.addElement(driver1);
        }
        this.letDriversReactToEachother(group1);
        group1 = new Vector();



        // south thru and south right
        group1 = new Vector();
        enum1 = ((Track) this.tracks().get("sThru")).drivers().elements();
        while(enum1.hasMoreElements()){
            Driver driver1 = (Driver) enum1.nextElement();

            group1.addElement(driver1);
        }
        enum2 = ((Track) this.tracks().get("sRight")).drivers().elements();
        while(enum2.hasMoreElements()){
            Driver driver1 = (Driver) enum2.nextElement();

            group1.addElement(driver1);
        }
        this.letDriversReactToEachother(group1);
        group1 = new Vector();



        // east thru and east right
        group1 = new Vector();
        enum1 = ((Track) this.tracks().get("eThru")).drivers().elements();
        while(enum1.hasMoreElements()){
            Driver driver1 = (Driver) enum1.nextElement();

            group1.addElement(driver1);
        }
        enum2 = ((Track) this.tracks().get("eRight")).drivers().elements();
        while(enum2.hasMoreElements()){
            Driver driver1 = (Driver) enum2.nextElement();

            group1.addElement(driver1);
        }
        this.letDriversReactToEachother(group1);
        group1 = new Vector();



        // west thru and west right
        group1 = new Vector();
        enum1 = ((Track) this.tracks().get("wThru")).drivers().elements();
        while(enum1.hasMoreElements()){
            Driver driver1 = (Driver) enum1.nextElement();

            group1.addElement(driver1);
        }
        enum2 = ((Track) this.tracks().get("wRight")).drivers().elements();
        while(enum2.hasMoreElements()){
            Driver driver1 = (Driver) enum2.nextElement();

            group1.addElement(driver1);
        }
        this.letDriversReactToEachother(group1);
        group1 = new Vector();




        // east left
        enum1 = ((Track) this.tracks().get("eLeft")).drivers().elements();
        while(enum1.hasMoreElements()){
            Driver driver1 = (Driver) enum1.nextElement();

            group1.addElement(driver1);
        }
        this.letDriversReactToEachother(group1);
        group1 = new Vector();



        // south left
        enum1 = ((Track) this.tracks().get("sLeft")).drivers().elements();
        while(enum1.hasMoreElements()){
            Driver driver1 = (Driver) enum1.nextElement();

            group1.addElement(driver1);
        }
        this.letDriversReactToEachother(group1);
        group1 = new Vector();



        // north left
        enum1 = ((Track) this.tracks().get("nLeft")).drivers().elements();
        while(enum1.hasMoreElements()){
            Driver driver1 = (Driver) enum1.nextElement();

            group1.addElement(driver1);
        }
        this.letDriversReactToEachother(group1);
        group1 = new Vector();

        // west left
        enum1 = ((Track) this.tracks().get("wLeft")).drivers().elements();
        while(enum1.hasMoreElements()){
            Driver driver1 = (Driver) enum1.nextElement();

            group1.addElement(driver1);
        }
        this.letDriversReactToEachother(group1);
        group1 = new Vector();









        // let drivers respond to signals
        Enumeration theOC = this.tracks().elements();
        while (theOC.hasMoreElements()) {
            Track aTrack = (Track) theOC.nextElement();
            aTrack.letDriversReact(this.view());
        }


    }
    private void sendSensorInputToTLC(){
        // send sensor input to TLC

        Enumeration theOC = this.tracks().elements();
        while (theOC.hasMoreElements()) {
            Track aTrack = (Track) theOC.nextElement();
            aTrack.sendSensorInputTo(this.view());
        }

    }

    private void giveCarsToIntersection(){
        // give the visible cars to the intersection canvas

        Enumeration theOC = this.tracks().elements();
        while (theOC.hasMoreElements()) {
            Track aTrack = (Track) theOC.nextElement();
            aTrack.handVisibleCarsTo(this.view());
        }

    }

//
//
// The following routines are just to write track data to disk
// This only needed to be done once, but we kept the code around
// just in case we needed it again.

    public double direction(TravelerFrame t2, TravelerFrame t1) {
        double angle = 0.0;
        boolean straight = false;
        int x2 = t2.position().x;
        int y2 = t2.position().y;
        int x1 = t1.position().x;
        int y1 = t1.position().y;
        double deltaX = x2 - x1;
        double deltaY = y2 - y1;

        if (x1 == x2) {
            // 0 or 180
            if (deltaY > 0.0) {
                return 180.0;
            }
            else {
                return 0.0;
            }
        }

        if (y1 == y2) {
            // 90 or 270
            if (deltaX > 0.0) {
                return 270.0;
            }
            else {
                return 90.0;
            }
        }

        double h = Math.sqrt((deltaX * deltaX) + (deltaY * deltaY));

        // handle the quadrant stuff

        if (deltaY > 0) {

            if (deltaX > 0) {
                // dy+ dx+
                angle = Math.asin(Math.abs(deltaX) / h) + (Math.PI);
            }
            else {
                // dy+ dx-
                angle = Math.asin(Math.abs(deltaY) / h) + (Math.PI / 2.0);
            }
        }
        else {
            if (deltaX > 0) {
                // dy- dx+
                angle = Math.asin(Math.abs(deltaY) / h) + (3.0 * Math.PI / 2.0);
            }
            else {
                // dy- dx-
                angle = Math.asin(Math.abs(deltaX) / h ) + (0.0);
            }
        }
      //  System.out.println("angle = " + angle*57.2958);
        return angle * 57.2958;
    }

    public String angleToImageKey(double angle) {
        int intValue = (new Double(angle)).intValue() / 10;
        intValue = intValue * 10;
        String stringAngle = Integer.toString(intValue);
        if (stringAngle.length() == 1) {
            return ("" + "00" + stringAngle);
        }
        if (stringAngle.length() == 2) {
            return ("" + "0" + stringAngle);
        }
        return "" + stringAngle;
    }

    public void writeTrackDataToDisk() {

        Enumeration theTS = this.tracks().elements();
        while (theTS.hasMoreElements()) {
            Track aTrack = (Track) theTS.nextElement();

            for (int i=0; i<8200; i++) {
                TravelerFrame theFrame = aTrack
                           .getAppropriateTravelerFrame(i);
                double theDirection = this.direction(aTrack
                           .getAppropriateTravelerFrame(i + 600), theFrame);

                theFrame.angle(this.angleToImageKey(theDirection));
            }
            CarImageFinder.writeToFile(aTrack.name(), aTrack.travelerFrames());
        }
    }

    public void letDriversReactToEachother(Vector v){
        if(v.size() <= 1) return;

        Enumeration driverGroup = v.elements();

        Driver guyInFront = new Driver();
        Driver aDriver = new Driver();
        Driver rememberThis = guyInFront;
        int dcpos = 5000;
        while( driverGroup.hasMoreElements() ){
            aDriver = (Driver) driverGroup.nextElement();
            int d1pos = aDriver.car().visibleCar().position();
            Enumeration driverGroup2 = v.elements();
            while( driverGroup2.hasMoreElements() ){
                Driver bDriver = (Driver) driverGroup2.nextElement();
                int d2pos = bDriver.car().visibleCar().position();
                if((d1pos > d2pos) & (d2pos < dcpos)){
                    guyInFront = bDriver;
                    dcpos = d2pos;
                }
            }
        }
        Driver driver1 = aDriver;
        Driver driver2 = guyInFront;
        if( guyInFront != rememberThis ){

            int driver1pos = driver1.car().visibleCar().position();
            int driver2pos = driver2.car().visibleCar().position();
//S.p("driver1pos = " + driver1pos);
//S.p("driver2pos = " + driver2pos);

            if( Math.abs( driver1pos-driver2pos ) < 1000 ){
                int driver1speed = driver1.car().currentVelocity();
                int driver2speed = driver2.car().currentVelocity();

                if( driver1pos < driver2pos ){
                    // driver2 is in front
                    driver1.car().acceleration(driver2pos - driver1pos-25 );

                }
                else{
                    // driver1 is in front
                    driver2.car().acceleration(driver1pos - driver2pos -25);

                }
                //S.p("driver1.car().acceleration() = " + driver1.car().acceleration());
                //S.p("driver2.car().acceleration() = " + driver2.car().acceleration());
            }
        }


    }


}
